==> org-lsst-ccs-ui-3.4.0-sources.jar: org/lsst/ccs/gconsole/plugins/monitor/AbstractMonitorView.java <==
package org.lsst.ccs.gconsole.plugins.monitor;

import java.util.*;
import java.util.AbstractMap.SimpleEntry;
import java.util.stream.Collectors;
import javax.swing.SwingUtilities;
import org.lsst.ccs.bus.data.AgentInfo;
import org.lsst.ccs.gconsole.base.filter.AgentChannelsFilter;
import org.lsst.ccs.gconsole.base.filter.PersistableAgentChannelsFilter;
import org.lsst.ccs.gconsole.services.aggregator.AgentChannel;
import org.lsst.ccs.gconsole.services.aggregator.AgentStatusEvent;

/**
 * An adapter that simplifies implementing {@code MonitorView}.
 *
 * @author onoprien
 */
abstract public class AbstractMonitorView implements PersistableMonitorView {

// -- Fields : -----------------------------------------------------------------
    
    protected PersistableAgentChannelsFilter filter;
    protected MonitorFormat formatter;
    
    /** Map of display paths to display channels, to be used by subclasses to populate the view. In addition order. */
    protected final LinkedHashMap<String,DisplayChannel> data = new LinkedHashMap<>();
    
    /** Policy on dealing with removed channels. */
    private volatile boolean discardRemovedChannels = false;
    
    private ArrayList<String> disconnectedAgents; // agents to be purged when they reconnect; null if empty.

// -- Life cycle : -------------------------------------------------------------
    
// -- Setters : ----------------------------------------------------------------

    /**
     * Sets the policy on dealing with channels reported as removed by the status aggregator.
     * If {@code true}, the channels are immediately removed from the {@code data} map.
     * If {@code false} (default), the channels are retained until and unless an identically
     * named agent appears in future status aggregator events.
     * 
     * @param discardRemovedChannels Policy value.
     */
    public void setDiscardRemovedChannels(boolean discardRemovedChannels) {
        this.discardRemovedChannels = discardRemovedChannels;
    }
    
    
// -- Implement AgentStatusListener : ------------------------------------------

    @Override
    public void connect(AgentStatusEvent event) {
        // ignore
    }

    @Override
    public void configure(AgentStatusEvent event) {
        if (!event.isEmpty()) {
            SwingUtilities.invokeLater(() -> onConnect(event));
        }
    }

    @Override
    public void statusChanged(AgentStatusEvent event) {
        if (!event.isEmpty()) {
            SwingUtilities.invokeLater(() -> onChange(event));
        }
    }

    @Override
    public void disconnect(AgentStatusEvent event) {
        if (!event.isEmpty()) {
            SwingUtilities.invokeLater(() -> onDisconnect(event));
        }
    }
    
    private void onConnect(AgentStatusEvent event) {
        List<AgentInfo> agents = event.getAgents();
        List<Map.Entry<String,AgentChannel>> channels = event.getAddedChannels().stream()
                .flatMap(channel -> filter.getDisplayPaths(channel).stream().map(displayPath -> new SimpleEntry<>(displayPath,channel)))
                .collect(Collectors.toList());
        addChannels(agents, channels);
    }
    
    private void onChange(AgentStatusEvent event) {
        List< Map.Entry<String,Map.Entry<AgentChannel,List<String>>> > channels = event.getStatusChanges().entrySet().stream()
                .flatMap(e -> filter.getDisplayPaths(e.getKey()).stream().map(displayPath -> new SimpleEntry<>(displayPath,e)))
                .collect(Collectors.toList());
        updateChannels(channels);
    }
    
    private void onDisconnect(AgentStatusEvent event) {
        List<AgentInfo> agents = event.getAgents();
        List<Map.Entry<String,AgentChannel>> channels = event.getRemovedChannels().stream()
                .flatMap(channel -> filter.getDisplayPaths(channel).stream().map(displayPath -> new SimpleEntry<>(displayPath,channel)))
                .collect(Collectors.toList());
        removeChannels(agents, channels);
    }
    
    
// -- Level 1 SPI for subclasses (called on EDT) : -----------------------------
    
    /**
     * Adds channels to this view.
     * Any previously existing channels from the same agents are removed.
     * 
     * @param agents Newly connected agents these channels belong to.
     * @param channels New channels, each entry contains display path and channel.
     */
    public void addChannels(List<AgentInfo> agents, List<Map.Entry<String,AgentChannel>> channels) {
        boolean needReset = doAdd(channels);
        if (!discardRemovedChannels) {
            needReset = purgeOfflineChannels(agents) || needReset;
        }
        if (needReset) {
            resetChannels();
        } else {
            update();
        }        
    }
    
    /**
     * Updates channels in this view.
     * 
     * @param channels Updated channels, each element contains: {display path : {channel : list of modified attributes}}.
     */
    public void updateChannels(List< Map.Entry<String,Map.Entry<AgentChannel,List<String>>> > channels) {
        doUpdate(channels);
    }
    
    /**
     * Removes channels from this view.
     * If the {@code discardRemovedChannels} flag is not set, the channels are updated but not removed.
     * 
     * @param agents Names of disconnected agents.
     * @param channels Removed channels, each entry contains display path and channel.
     */
    public void removeChannels(List<AgentInfo> agents, List<Map.Entry<String,AgentChannel>> channels) {
        List<String> agentNames = agents.stream().map(a -> a.getName()).collect(Collectors.toList());
        if (discardRemovedChannels) {
            if (doRemove(agentNames)) {
                resetChannels();
            } else {
                update();
            }
        } else {
            if (disconnectedAgents == null) {
                disconnectedAgents = new ArrayList<>(agentNames);
            } else {
                disconnectedAgents.addAll(agentNames);
            }
            doUpdate(channels.stream()
                    .map(e -> new SimpleEntry<String, Map.Entry<AgentChannel, List<String>>>( e.getKey(), new SimpleEntry<AgentChannel,List<String>>(e.getValue(), null) ))
                    .collect(Collectors.toList()));
        }
    }
    
    
// -- Local methods used for going from stage 1 to stage 2 : -------------------
    
    /**
     * Adds channels to {@code data}, replacing existing ones and creating new display channels if necessary.
     * 
     * @param channels List of {display path : AgentChannel}.
     * @return True if new display channels have been created.
     */
    private boolean doAdd(List<Map.Entry<String,AgentChannel>> channels) {
        boolean needReset = false;
        for (Map.Entry<String,AgentChannel> e : channels) {
            String dp = e.getKey();
            AgentChannel c = e.getValue();
            DisplayChannel dc = data.get(dp);
            if (dc == null) {
                dc = createChannel(dp);
                dc.addChannel(c);
                data.put(dp, dc);
                needReset = true;
            } else {
                dc.addChannel(c);
                dc.update(null);
            }
        }
        return needReset;
    }
    
    /**
     * Removes channels in OFFLINE state from {@code data}.
     * 
     * @param agents Agents to check.
     * @return 
     */
    private boolean purgeOfflineChannels(List<AgentInfo> agents) {
        
        if (disconnectedAgents == null) return false;
        
        Set<String> eventAgents = agents.stream().map(a -> a.getName()).collect(Collectors.toSet());
        ArrayList<String> agentsToPurge = new ArrayList<>(eventAgents.size());
        Iterator<String> it = disconnectedAgents.iterator();
        while (it.hasNext()) {
            String a = it.next();
            if (eventAgents.contains(a)) {
                agentsToPurge.add(a);
                it.remove();
            }
        }
        if (disconnectedAgents.isEmpty()) disconnectedAgents = null;
        if (agentsToPurge.isEmpty()) return false;
        
        boolean needReset = false;
        Iterator<Map.Entry<String, DisplayChannel>> itt = data.entrySet().iterator();
        while (itt.hasNext()) {
            Map.Entry<String, DisplayChannel> e = itt.next();
            if (e.getValue().purgeOfflineChannels(agentsToPurge)) {
                itt.remove();
                needReset = true;
            }
        }
        
        return needReset;
    }
    
    /**
     * Removes data channels that belong to the specified agents from {@code data}.
     * 
     * @param agents Agents to remove.
     * @return True if display channels have been removed from {@code data}.
     */    
    private boolean doRemove(List<String> agents) {
        boolean needReset = false;
        Iterator<Map.Entry<String,DisplayChannel>> it = data.entrySet().iterator();
        while (it.hasNext()) {
            DisplayChannel dc = it.next().getValue();
            List<AgentChannel> oldChannels = dc.getChannels();
            if (oldChannels.stream().anyMatch(a -> agents.contains(a.getAgentName()))) {
                ArrayList<AgentChannel> newChannels = new ArrayList<>(oldChannels.size());
                for (AgentChannel c : oldChannels) {
                    if (!agents.contains(c.getAgentName())) {
                        newChannels.add(c);
                    }
                }
                if (newChannels.isEmpty()) {
                    it.remove();
                    needReset = true;
                } else {
                    dc.setChannels(newChannels);
                    dc.update(null);
                }
            }
        }
        return needReset;
    }
    
    private void doUpdate(List< Map.Entry<String,Map.Entry<AgentChannel,List<String>>> > channels) {
        for (Map.Entry<String,Map.Entry<AgentChannel,List<String>>> e : channels) {
            String displayPath = e.getKey();
            DisplayChannel dc = data.get(displayPath);
            List<String> attributes = e.getValue().getValue();
            if (dc != null) dc.update(attributes);
        }
        update();
    }

    
// -- Level 2 SPI for subclasses (called on EDT) : -----------------------------
    
    /**
     * Called on EDT whenever channels have been added or removed.
     * Should be implemented to completely rebuild the view based on {@code data} field.
     * Empty implementation is provided.
     */
    protected void resetChannels() {}
    
    /**
     * Called on EDT at the end of each update, after all DisplayChannel.update(...) methods.
     * Empty implementation is provided.
     */
    protected void update() {}
    
    /**
     * Called on EDT to create new display channels whenever this view is informed of new display paths.
     * The default implementation constructs an instance of {@link DisplayChannelMulti} without any {@link Updatable} target.
     * 
     * @param displayPath Display path.
     * @return An instance of {@code DisplayChannel} to be put into the {@code data} map for the specified display path key.
     */
    protected DisplayChannel createChannel(String displayPath) {
        return new DisplayChannelMulti(displayPath);
    }


// -- Implement MonitorView : --------------------------------------------------
    
    /**
     * Returns {@code true} if this view has no channels to display.
     * Implemented to return {@code true} if the {@code data} map is empty.
     * 
     * @return True if this view has no display channels.
     */
    @Override
    public boolean isEmpty() {
        return data.isEmpty();
    }

    @Override
    public List<String> getGroups() {
        List<String> out = PersistableMonitorView.super.getGroups();
        if (out == null) {
            LinkedHashSet<String> groups = new LinkedHashSet<>();
            data.keySet().forEach(displayPath -> {
                String group = getGroup(displayPath);
                if (group != null) groups.add(group);
            });
            return new ArrayList<>(groups);
        }
        return out;
    }

    @Override
    public PersistableAgentChannelsFilter getFilter() {
        return filter;
    }

    @Override
    public void setFilter(AgentChannelsFilter filter) {
        if (!(filter instanceof PersistableAgentChannelsFilter)) throw new IllegalArgumentException(getClass().getName() +" only accepts AbstractChannelsFilter filters");
        this.filter = (PersistableAgentChannelsFilter) filter;
    }

    @Override
    public MonitorFormat getFormater() {
        return formatter;
    }

    @Override
    public void setFormatter(MonitorFormat formatter) {
        this.formatter = formatter;
    }

}

==> org-lsst-ccs-ui-3.4.4-sources.jar: org/lsst/ccs/gconsole/plugins/monitor/AbstractMonitorView.java <==
package org.lsst.ccs.gconsole.plugins.monitor;

import java.util.*;
import java.util.AbstractMap.SimpleEntry;
import java.util.stream.Collectors;
import javax.swing.SwingUtilities;
import org.lsst.ccs.bus.data.AgentInfo;
import org.lsst.ccs.gconsole.base.filter.AgentChannelsFilter;
import org.lsst.ccs.gconsole.base.filter.PersistableAgentChannelsFilter;
import org.lsst.ccs.gconsole.services.aggregator.AgentChannel;
import org.lsst.ccs.gconsole.services.aggregator.AgentStatusEvent;

/**
 * An adapter that simplifies implementing {@code MonitorView}.
 *
 * @author onoprien
 */
abstract public class AbstractMonitorView implements PersistableMonitorView {

// -- Fields : -----------------------------------------------------------------
    
    protected PersistableAgentChannelsFilter filter;
    protected MonitorFormat formatter;
    
    /** Map of display paths to display channels, to be used by subclasses to populate the view. In addition order. */
    protected final LinkedHashMap<String,DisplayChannel> data = new LinkedHashMap<>();
    
    /** Policy on dealing with removed channels. */
    private volatile boolean discardRemovedChannels = false;
    
    private ArrayList<String> disconnectedAgents; // agents to be purged when they reconnect; null if empty.

// -- Life cycle : -------------------------------------------------------------
    
// -- Setters : ----------------------------------------------------------------

    /**
     * Sets the policy on dealing with channels reported as removed by the status aggregator.
     * If {@code true}, the channels are immediately removed from the {@code data} map.
     * If {@code false} (default), the channels are retained until and unless an identically
     * named agent appears in future status aggregator events.
     * 
     * @param discardRemovedChannels Policy value.
     */
    public void setDiscardRemovedChannels(boolean discardRemovedChannels) {
        this.discardRemovedChannels = discardRemovedChannels;
    }
    
    
// -- Implement AgentStatusListener : ------------------------------------------

    @Override
    public void connect(AgentStatusEvent event) {
        // ignore
    }

    @Override
    public void configure(AgentStatusEvent event) {
        if (!event.isEmpty()) {
            SwingUtilities.invokeLater(() -> onConnect(event));
        }
    }

    @Override
    public void statusChanged(AgentStatusEvent event) {
        if (!event.isEmpty()) {
            SwingUtilities.invokeLater(() -> onChange(event));
        }
    }

    @Override
    public void disconnect(AgentStatusEvent event) {
        if (!event.isEmpty()) {
            SwingUtilities.invokeLater(() -> onDisconnect(event));
        }
    }
    
    private void onConnect(AgentStatusEvent event) {
        List<AgentInfo> agents = event.getAgents();
        List<Map.Entry<String,AgentChannel>> channels = event.getAddedChannels().stream()
                .flatMap(channel -> filter.getDisplayPaths(channel).stream().map(displayPath -> new SimpleEntry<>(displayPath,channel)))
                .collect(Collectors.toList());
        addChannels(agents, channels);
    }
    
    private void onChange(AgentStatusEvent event) {
        List< Map.Entry<String,Map.Entry<AgentChannel,List<String>>> > channels = event.getStatusChanges().entrySet().stream()
                .flatMap(e -> filter.getDisplayPaths(e.getKey()).stream().map(displayPath -> new SimpleEntry<>(displayPath,e)))
                .collect(Collectors.toList());
        updateChannels(channels);
    }
    
    private void onDisconnect(AgentStatusEvent event) {
        List<AgentInfo> agents = event.getAgents();
        List<Map.Entry<String,AgentChannel>> channels = event.getRemovedChannels().stream()
                .flatMap(channel -> filter.getDisplayPaths(channel).stream().map(displayPath -> new SimpleEntry<>(displayPath,channel)))
                .collect(Collectors.toList());
        removeChannels(agents, channels);
    }
    
    
// -- Level 1 SPI for subclasses (called on EDT) : -----------------------------
    
    /**
     * Adds channels to this view.
     * Any previously existing channels from the same agents are removed.
     * 
     * @param agents Newly connected agents these channels belong to.
     * @param channels New channels, each entry contains display path and channel.
     */
    public void addChannels(List<AgentInfo> agents, List<Map.Entry<String,AgentChannel>> channels) {
        boolean needReset = false;
        LinkedHashSet<DisplayChannel> updatedChannels = new LinkedHashSet<>();
        
        // Remove channels from added agents if they were not discarded earlier
        
        if (!discardRemovedChannels) {
            updatedChannels = doRemove(agents.stream().map(a -> a.getName()).collect(Collectors.toList()), false);
        }
        
        // Add agents

        for (Map.Entry<String,AgentChannel> e : channels) {
            String dp = e.getKey();
            AgentChannel c = e.getValue();
            DisplayChannel dc = data.get(dp);
            if (dc == null) {
                dc = createChannel(dp);
                data.put(dp, dc);
                needReset = true;
            } else {
                updatedChannels.add(dc);
            }
            dc.appendChannel(c);
        }
        
        // Remove display channels that have no agent channels
        
        Iterator<DisplayChannel> it = updatedChannels.iterator();
        while (it.hasNext()) {
            DisplayChannel dc = it.next();
            if (dc.getChannels().isEmpty()) {
                data.remove(dc.getPath());
                it.remove();
                needReset = true;
            }
        }
        
        // Call SPI hooks
        
        if (needReset) {
            resetChannels();
        } else {
            for (DisplayChannel dc : updatedChannels) {
                dc.update(null);
            }
            update();
        }        
    }
    
    /**
     * Updates channels in this view.
     * 
     * @param channels Updated channels, each element contains: {display path : {channel : list of modified attributes}}.
     */
    public void updateChannels(List< Map.Entry<String,Map.Entry<AgentChannel,List<String>>> > channels) {
        doUpdate(channels);
    }
    
    /**
     * Removes channels from this view.
     * If the {@code discardRemovedChannels} flag is not set, the channels are updated but not removed.
     * 
     * @param agents Names of disconnected agents.
     * @param channels Removed channels, each entry contains display path and channel.
     */
    public void removeChannels(List<AgentInfo> agents, List<Map.Entry<String,AgentChannel>> channels) {
        List<String> agentNames = agents.stream().map(a -> a.getName()).collect(Collectors.toList());
        if (discardRemovedChannels) {
            if (!doRemove(agentNames, true).isEmpty()) {
                resetChannels();
            } else {
                update();
            }
        } else {
            if (disconnectedAgents == null) {
                disconnectedAgents = new ArrayList<>(agentNames);
            } else {
                disconnectedAgents.addAll(agentNames);
            }
            doUpdate(channels.stream()
                    .map(e -> new SimpleEntry<String, Map.Entry<AgentChannel, List<String>>>( e.getKey(), new SimpleEntry<AgentChannel,List<String>>(e.getValue(), null) ))
                    .collect(Collectors.toList()));
        }
    }
    
    
// -- Local methods used for going from stage 1 to stage 2 : -------------------
    
    /**
     * Removes data channels that belong to the specified agents from {@code data}.
     * 
     * @param agents Agents to remove.
     * @return True if display channels have been removed from {@code data}.
     */    
    private LinkedHashSet<DisplayChannel> doRemove(List<String> agents, boolean purge) {
        LinkedHashSet<DisplayChannel> out = new LinkedHashSet<>(data.size()*2);
        Iterator<Map.Entry<String,DisplayChannel>> it = data.entrySet().iterator();
        while (it.hasNext()) {
            DisplayChannel dc = it.next().getValue();
            List<AgentChannel> oldChannels = dc.getChannels();
            if (oldChannels.stream().anyMatch(a -> agents.contains(a.getAgentName()))) {
                ArrayList<AgentChannel> newChannels = new ArrayList<>(oldChannels.size());
                for (AgentChannel c : oldChannels) {
                    if (!agents.contains(c.getAgentName())) {
                        newChannels.add(c);
                    }
                }
                dc.setChannels(newChannels);
                if (newChannels.isEmpty()) {
                    out.add(dc);
                    if (purge) it.remove();
                } else {
                    if (purge) dc.update(null);
                }
            }
        }
        return out;
    }
    
    private void doUpdate(List< Map.Entry<String,Map.Entry<AgentChannel,List<String>>> > channels) {
        for (Map.Entry<String,Map.Entry<AgentChannel,List<String>>> e : channels) {
            String displayPath = e.getKey();
            DisplayChannel dc = data.get(displayPath);
            List<String> attributes = e.getValue().getValue();
            if (dc != null) dc.update(attributes);
        }
        update();
    }

    
// -- Level 2 SPI for subclasses (called on EDT) : -----------------------------
    
    /**
     * Called on EDT whenever channels have been added or removed.
     * Should be implemented to completely rebuild the view based on {@code data} field.
     * Empty implementation is provided.
     */
    protected void resetChannels() {}
    
    /**
     * Called on EDT at the end of each update, after all DisplayChannel.update(...) methods.
     * Empty implementation is provided.
     */
    protected void update() {}
    
    /**
     * Called on EDT to create new display channels whenever this view is informed of new display paths.
     * The default implementation constructs an instance of {@link DisplayChannelMulti} without any {@link Updatable} target.
     * 
     * @param displayPath Display path.
     * @return An instance of {@code DisplayChannel} to be put into the {@code data} map for the specified display path key.
     */
    protected DisplayChannel createChannel(String displayPath) {
        return new DisplayChannelMulti(displayPath);
    }


// -- Implement MonitorView : --------------------------------------------------
    
    /**
     * Returns {@code true} if this view has no channels to display.
     * Implemented to return {@code true} if the {@code data} map is empty.
     * 
     * @return True if this view has no display channels.
     */
    @Override
    public boolean isEmpty() {
        return data.isEmpty();
    }

    @Override
    public List<String> getGroups() {
        List<String> out = PersistableMonitorView.super.getGroups();
        if (out == null) {
            LinkedHashSet<String> groups = new LinkedHashSet<>();
            data.keySet().forEach(displayPath -> {
                String group = getGroup(displayPath);
                if (group != null) groups.add(group);
            });
            return new ArrayList<>(groups);
        }
        return out;
    }

    @Override
    public PersistableAgentChannelsFilter getFilter() {
        return filter;
    }

    @Override
    public void setFilter(AgentChannelsFilter filter) {
        if (!(filter instanceof PersistableAgentChannelsFilter)) throw new IllegalArgumentException(getClass().getName() +" only accepts AbstractChannelsFilter filters");
        this.filter = (PersistableAgentChannelsFilter) filter;
    }

    @Override
    public MonitorFormat getFormater() {
        return formatter;
    }

    @Override
    public void setFormatter(MonitorFormat formatter) {
        this.formatter = formatter;
    }

}
